"""CSV target sink class, which handles writing streams."""

from __future__ import annotations

import datetime
import functools
import sys
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytz
from singer_sdk.sinks import BatchSink

from meltanolabs_target_csv.serialization import write_batch, write_header

if TYPE_CHECKING:
    from singer_sdk import Target


class CSVSink(BatchSink):
    """CSV target sink class."""

    MAX_SIZE_DEFAULT = sys.maxsize

    def __init__(  # noqa: D107
        self,
        target: Target,
        stream_name: str,
        schema: dict,
        key_properties: list[str] | None,
    ) -> None:
        self._timestamp_time: datetime.datetime | None = None
        super().__init__(target, stream_name, schema, key_properties)

    @property
    def timestamp_time(self) -> datetime.datetime:  # noqa: D102
        if not self._timestamp_time:
            self._timestamp_time = datetime.datetime.now(
                tz=pytz.timezone(self.config["timestamp_timezone"])
            )

        return self._timestamp_time

    @property
    def filepath_replacement_map(self) -> dict[str, str]:  # noqa: D102
        return {
            "stream_name": self.stream_name,
            "datestamp": self.timestamp_time.strftime(self.config["datestamp_format"]),
            "timestamp": self.timestamp_time.strftime(self.config["timestamp_format"]),
        }

    @property
    def output_file(self) -> Path:  # noqa: D102
        filename = self.config["file_naming_scheme"]
        for key, val in self.filepath_replacement_map.items():
            replacement_pattern = f"{{{key}}}"
            if replacement_pattern in filename:
                filename = filename.replace(replacement_pattern, val)

        if "output_path_prefix" in self.config:
            warnings.warn(
                "The property `output_path_prefix` is deprecated, "
                "please use `output_path`.",
                category=UserWarning,
            )

        # Accept all possible properties defining the output path.
        # - output_path: The new designated property.
        # - destination_path: Alias for `output_path` (`hotgluexyz` compat).
        # - output_path_prefix: The property used up until now.
        output_path = self.config.get(
            "output_path",
            self.config.get(
                "destination_path", self.config.get("output_path_prefix", None)
            ),
        )

        filepath = Path(filename)
        if output_path is not None:
            filepath = Path(output_path) / filepath

        return filepath

    @functools.cached_property
    def keys(self) -> list[str]:
        """Get the header keys for the CSV file."""
        if "properties" not in self.schema:
            raise ValueError("Stream's schema has no properties defined")

        return list(self.schema["properties"].keys())

    @functools.cached_property
    def escape_character(self) -> str | None:
        """Get the escape character for the CSV file."""
        return self.config.get("escape_character")

    @functools.cached_property
    def encoding(self) -> str:
        """Get the encoding for the CSV file."""
        return self.config.get("encoding", "utf-8")

    def setup(self) -> None:
        """Create the output file and write the header."""
        super().setup()
        output_file = self.output_file
        self.logger.info("Writing to destination file '%s'...", output_file.resolve())
        write_header(
            output_file,
            self.keys,
            dialect="excel",
            escapechar=self.escape_character,
            encoding=self.encoding,
        )

    def process_batch(self, context: dict) -> None:
        """Write out any prepped records and return once fully written."""
        output_file: Path = self.output_file

        if not isinstance(context["records"], list):
            self.logger.warning("No values in %s records collection.", self.stream_name)
            context["records"] = []

        records: list[dict[str, Any]] = context["records"]
        if "record_sort_property_name" in self.config:
            sort_property_name = self.config["record_sort_property_name"]
            records = sorted(records, key=lambda x: x[sort_property_name])

        self.logger.info(f"Appending {len(records)} records to file...")

        write_batch(
            output_file,
            records,
            self.keys,
            dialect="excel",
            escapechar=self.escape_character,
            encoding=self.encoding,
        )



"""Serialization utilities for CSV files."""

from __future__ import annotations

import csv  # noqa: D100
import tempfile
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


def write_csv(
    filepath: Path,
    records: list[dict],
    keys: list[str],
    **kwargs: Any,
) -> int:
    """Write a CSV file."""
    with open(filepath, "w", encoding="utf-8", newline="") as fp:
        writer = csv.DictWriter(fp, fieldnames=keys, dialect="excel", **kwargs)
        writer.writeheader()
        for record_count, record in enumerate(records, start=1):
            writer.writerow(record)

    return record_count


def write_header(
    filepath: Path, keys: list[str], encoding: str = "utf-8", **kwargs: Any
) -> None:
    """Write a header to a CSV file.

    Creates the parent directory if it doesn't exist.

    Args:
        filepath: Path to the CSV file
        keys: List of column names
        encoding: Character encoding to use (default: utf-8)
        **kwargs: Additional arguments to pass to csv.DictWriter
    """
    filepath.parent.mkdir(parents=True, exist_ok=True)
    with filepath.open("w", encoding=encoding, newline="") as fp:
        writer = csv.DictWriter(fp, fieldnames=keys, **kwargs)
        writer.writeheader()


def write_batch(
    filepath: Path,
    records: list[dict],
    keys: list[str],
    encoding: str = "utf-8",
    **kwargs: Any,
) -> None:
    """Write a batch of records to a CSV file.

    Args:
        filepath: Path to the CSV file
        records: List of record dictionaries to write
        keys: List of column names
        encoding: Character encoding to use (default: utf-8)
        **kwargs: Additional arguments to pass to csv.DictWriter
    """
    with tempfile.NamedTemporaryFile("w+", encoding=encoding, newline="") as tmp_fp:
        writer = csv.DictWriter(tmp_fp, fieldnames=keys, **kwargs)
        writer.writerows(records)

        tmp_fp.seek(0)

        with filepath.open("a", encoding=encoding) as f:
            f.write(tmp_fp.read())


def read_csv(filepath: Path) -> list[dict]:
    """Read a CSV file."""
    result: list[dict] = []
    with open(filepath, newline="") as fp:
        reader = csv.DictReader(fp, delimiter=",", dialect="excel")
        result.extend(iter(reader))
    return result



"""CSV target class."""

from singer_sdk import typing as th
from singer_sdk.target_base import Target

from meltanolabs_target_csv.sinks import CSVSink


class TargetCSV(Target):
    """A Singer target that generates CSV files."""

    name = "target-csv"
    package_name = "meltanolabs-target-csv"

    config_jsonschema = th.PropertiesList(
        th.Property(
            "output_path",
            th.StringType,
            description=(
                "Filesystem path where to store output files. "
                "By default, the current working directory will be used."
            ),
        ),
        th.Property(
            "destination_path",
            th.StringType,
            description=(
                "Filesystem path where to store output files. Alias for "
                "`output_path` to be compatible with the `hotgluexyz` variant."
            ),
        ),
        th.Property(
            "output_path_prefix",
            th.StringType,
            description=("DEPRECATED. Filesystem path where to store output files."),
        ),
        th.Property(
            "file_naming_scheme",
            th.StringType,
            description=(
                "The scheme with which output files will be named. "
                "Naming scheme may leverage any of the following substitutions: \n\n"
                "- `{stream_name}`"
                "- `{datestamp}`"
                "- `{timestamp}`"
            ),
            default="{stream_name}.csv",
        ),
        th.Property(
            "datestamp_format",
            th.StringType,
            description=(
                "A python format string to use when outputting the `{datestamp}` "
                "string. For reference, see: "
                "https://docs.python.org/3/library/datetime.html"
                "#strftime-and-strptime-format-codes"
            ),
            default="%Y-%m-%d",
        ),
        th.Property(
            "timestamp_format",
            th.StringType,
            description=(
                "A python format string to use when outputting the `{timestamp}` "
                "string. For reference, see: "
                "https://docs.python.org/3/library/datetime.html"
                "#strftime-and-strptime-format-codes"
            ),
            default="%Y-%m-%d.T%H%M%S",
        ),
        th.Property(
            "timestamp_timezone",
            th.StringType,
            description=(
                "The timezone code or name to use when generating "
                "`{timestamp}` and `{datestamp}`. "
                "Defaults to 'UTC'. For a list of possible values, please see: "
                "https://en.wikipedia.org/wiki/List_of_tz_database_time_zones"
            ),
            default="UTC",
        ),
        th.Property(
            "stream_maps",
            th.ObjectType(),
            description=(
                "Allows inline stream transformations and aliasing. "
                "For more information see: "
                "https://sdk.meltano.com/en/latest/stream_maps.html"
            ),
        ),
        th.Property(
            "record_sort_property_name",
            th.StringType,
            description=(
                "A property in the record which will be used as a sort key.\n\n"
                "If this property is omitted, records will not be sorted."
            ),
        ),
        th.Property(
            "overwrite_behavior",
            th.StringType,
            description=(
                "Determines the overwrite behavior if destination file already exists. "
                "Must be one of the following string values: \n\n"
                "- `append_records` (default) - append records at the insertion point\n"
                "- `replace_file` - replace entire file using `default_CSV_template`\n"
            ),
            default="replace_file",
        ),
        th.Property(
            "escape_character",
            th.StringType(min_length=1, max_length=1),
            description="The character to use for escaping special characters.",
        ),
        th.Property(
            "encoding",
            th.StringType,
            description=(
                "The character encoding to use when writing CSV files. "
                "Defaults to 'utf-8'. For a list of possible values, see: "
                "https://docs.python.org/3/library/codecs.html#standard-encodings"
            ),
            default="utf-8",
        ),
    ).to_dict()
    default_sink_class = CSVSink



"""Tests standard target features using the built-in SDK tests library."""

from __future__ import annotations

from importlib.resources import files
from typing import TYPE_CHECKING, Any

from singer_sdk.testing import get_target_test_class
from singer_sdk.testing.suites import SingerTestSuite
from singer_sdk.testing.templates import TargetFileTestTemplate

from meltanolabs_target_csv.target import TargetCSV

from . import data_files

if TYPE_CHECKING:
    from importlib.abc import Traversable

SAMPLE_CONFIG: dict[str, Any] = {
    "escape_character": "\\",
}


class MultipleStreamsTest(TargetFileTestTemplate):
    name = "users_and_employees"

    @property
    def singer_filepath(self) -> Traversable:
        return files(data_files) / f"{self.name}.singer"


TestTargetCSV = get_target_test_class(
    TargetCSV,
    config=SAMPLE_CONFIG,
    custom_suites=[
        SingerTestSuite(
            kind="target",
            tests=[MultipleStreamsTest],
        )
    ],
)



from pathlib import Path

import pytest

from meltanolabs_target_csv.sinks import CSVSink
from meltanolabs_target_csv.target import TargetCSV


def test_sink_output_file_cwd():
    """Verify `output_file` property without defining an output path."""
    target = TargetCSV()
    sink = CSVSink(
        target=target, stream_name="foo", schema={"properties": {}}, key_properties=[]
    )
    assert sink.output_file == Path("foo.csv")


@pytest.mark.parametrize(
    "property_name", ["output_path", "destination_path", "output_path_prefix"]
)
def test_sink_output_file_with_path(tmp_path, property_name):
    """Verify `output_file` property when defining an output path.

    The test is parameterized to iterate and verify all the possible properties
    which define an output path.
    """
    folder_path = tmp_path / "to/folder"
    output_file = tmp_path / "to/folder/foo.csv"
    target = TargetCSV(config={property_name: str(folder_path)})
    sink = CSVSink(
        target=target, stream_name="foo", schema={"properties": {}}, key_properties=[]
    )
    assert sink.output_file == output_file


def test_sink_output_file_with_path_deprecated(tmp_path):
    """Verify `output_file` property with deprecated `output_path_prefix` property."""
    folder_path = tmp_path / "to/folder"
    output_file = tmp_path / "to/folder/foo.csv"
    target = TargetCSV(config={"output_path_prefix": str(folder_path)})
    sink = CSVSink(
        target=target, stream_name="foo", schema={"properties": {}}, key_properties=[]
    )
    with pytest.warns(
        UserWarning,
        match="The property `output_path_prefix` is deprecated, "
        "please use `output_path`.",
    ):
        assert sink.output_file == output_file



"""CSV-related tests."""

from __future__ import annotations

from pathlib import Path
from typing import Any

import pytest

from meltanolabs_target_csv.serialization import read_csv, write_batch, write_header

SAMPLE_DATASETS: list[tuple[dict, list[dict[str, Any]]]] = [
    (
        # JSON Schema
        {
            "properties": {
                "int_key": {"type": "int"},
                "complex_field": {"type": ["object", "null"]},
                "string_field": {"type": "string"},
            }
        },
        # Records
        [
            {"int_key": 1, "string_field": "hello", "complex_field": {}},
            {"int_key": 2, "complex_field": {"foo": "bar"}, "string_field": "hello"},
            # Fields out of order:
            {
                "int_key": 3,
                "string_field": "hello",
                "complex_field": {"foo": "bar"},
            },
            # Special characters in string:
            {"int_key": 4, "string_field": "''he\\l|,o\"\\n", "complex_field": {}},
            # Newline characters in string:
            {
                "int_key": 5,
                "complex_field": {},
                "string_field": "hello\nworld",
            },
        ],
    )
]


@pytest.fixture
def output_dir() -> Path:
    result = Path("./.output/")
    if not result.is_dir():
        result.mkdir()

    return result


@pytest.fixture
def output_filepath(output_dir) -> Path:
    result = Path(output_dir / "csv-test-output.csv")
    if result.exists():
        result.unlink()

    return result


@pytest.fixture
def test_file_paths(output_dir) -> list[Path]:
    paths = []
    for dir in range(4):
        path = Path(output_dir / f"test-dir-{dir}/csv-test-output-{dir}.csv")
        path.unlink(missing_ok=True)
        paths.append(path)
    return paths


def test_csv_write(output_filepath) -> None:
    for schema, records in SAMPLE_DATASETS:
        keys = list(schema["properties"].keys())
        write_header(filepath=output_filepath, keys=keys)
        write_batch(filepath=output_filepath, records=records, keys=keys)


def test_csv_write_if_not_exists(test_file_paths) -> None:
    for path in test_file_paths:
        for schema, records in SAMPLE_DATASETS:
            keys = list(schema["properties"].keys())
            write_header(filepath=path, keys=keys)
            write_batch(filepath=path, records=records, keys=keys)


def test_csv_roundtrip(output_filepath) -> None:
    for schema, records in SAMPLE_DATASETS:
        keys = list(schema["properties"].keys())
        write_header(filepath=output_filepath, keys=keys)
        write_batch(filepath=output_filepath, records=records, keys=keys)
        read_records = read_csv(filepath=output_filepath)
        for orig_record, new_record in zip(records, read_records):
            for key in orig_record:
                # Note: Results are stringified during serialization
                assert str(orig_record[key]) == new_record[key]
